@@ -24,9 +24,7 @@
  *	Peter Hutterer (peter.hutterer@redhat.com)
  */
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #include "libwacomint.h"
 #include <assert.h>
@@ -38,6 +36,10 @@
 
 #include <linux/input-event-codes.h>
 
+#if !HAVE_G_MEMDUP2
+#define g_memdup2 g_memdup
+#endif
+
 static const WacomDevice *
 libwacom_get_device(const WacomDeviceDatabase *db, const char *match)
 {
@@ -172,7 +174,7 @@ get_bus (GUdevDevice *device)
 	parent = g_object_ref (device);
 
 	while (parent && subsystem &&
-	       (streq(subsystem, "input") || streq (subsystem, "hid"))) {
+	       (g_str_equal(subsystem, "input") || g_str_equal (subsystem, "hid"))) {
 		GUdevDevice *old_parent = parent;
 		parent = g_udev_device_get_parent (old_parent);
 		if (parent)
@@ -181,7 +183,7 @@ get_bus (GUdevDevice *device)
 	}
 
 	if (parent) {
-		if (subsystem && (streq(subsystem, "tty") || streq(subsystem, "serio")))
+		if (subsystem && (g_str_equal(subsystem, "tty") || g_str_equal(subsystem, "serio")))
 			bus_str = g_strdup ("serial");
 		else
 			bus_str = g_strdup (subsystem);
@@ -193,6 +195,24 @@ get_bus (GUdevDevice *device)
 	return bus_str;
 }
 
+static GUdevDevice *
+client_query_by_subsystem_and_device_file (GUdevClient *client,
+					   const char  *subsystem,
+					   const char  *path)
+{
+	GList *l, *devices;
+	GUdevDevice *ret = NULL;
+
+	devices = g_udev_client_query_by_subsystem (client, subsystem);
+	for (l = devices; l != NULL; l = l->next) {
+		if (!ret && g_strcmp0 (g_udev_device_get_device_file (l->data), path) == 0)
+			ret = g_object_ref (l->data);
+		g_object_unref (l->data);
+	}
+	g_list_free (devices);
+	return ret;
+}
+
 static gboolean
 get_device_info (const char            *path,
 		 int                   *vendor_id,
@@ -215,7 +235,9 @@ get_device_info (const char            *path,
 	*name = NULL;
 	bus_str = NULL;
 	client = g_udev_client_new (subsystems);
-	device = g_udev_client_query_by_device_file (client, path);
+	device = client_query_by_subsystem_and_device_file (client, subsystems[0], path);
+	if (device == NULL)
+		device = g_udev_client_query_by_device_file (client, path);
 	if (device == NULL) {
 		libwacom_error_set(error, WERROR_INVALID_PATH, "Could not find device '%s' in udev", path);
 		goto out;
@@ -309,7 +331,8 @@ static WacomDevice *
 libwacom_copy(const WacomDevice *device)
 {
 	WacomDevice *d;
-	int i;
+	GHashTableIter iter;
+	gpointer k, v;
 
 	d = g_new0 (WacomDevice, 1);
 	g_atomic_int_inc(&d->refcnt);
@@ -319,12 +342,14 @@ libwacom_copy(const WacomDevice *device)
 	d->height = device->height;
 	d->integration_flags = device->integration_flags;
 	d->layout = g_strdup(device->layout);
-	d->nmatches = device->nmatches;
-	d->matches = g_malloc((d->nmatches + 1) * sizeof(WacomMatch*));
-	for (i = 0; i < d->nmatches; i++)
-		d->matches[i] = libwacom_match_ref(device->matches[i]);
-	d->matches[d->nmatches] = NULL;
-	d->match = device->match;
+	d->matches = g_array_sized_new(TRUE, TRUE, sizeof(WacomDevice*),
+				       device->matches->len);
+	for (guint i = 0; i < device->matches->len; i++) {
+		WacomMatch *m = g_array_index(device->matches, WacomMatch*, i);
+		libwacom_match_ref(m);
+		g_array_append_val(d->matches, m);
+	}
+	d->match = libwacom_match_ref(device->match);
 	if (device->paired)
 		d->paired = libwacom_match_ref(device->paired);
 	d->cls = device->cls;
@@ -333,19 +358,39 @@ libwacom_copy(const WacomDevice *device)
 	d->strips_num_modes = device->strips_num_modes;
 	d->ring_num_modes = device->ring_num_modes;
 	d->ring2_num_modes = device->ring2_num_modes;
-	d->num_styli = device->num_styli;
-	d->supported_styli = g_memdup (device->supported_styli, sizeof(int) * device->num_styli);
-	d->num_leds = device->num_leds;
-	d->status_leds = g_memdup (device->status_leds, sizeof(WacomStatusLEDs) * device->num_leds);
-	d->num_buttons = device->num_buttons;
-	d->buttons = g_memdup (device->buttons, sizeof(WacomButtonFlags) * device->num_buttons);
-	d->button_codes = g_memdup (device->button_codes, sizeof(int) * device->num_buttons);
+	d->styli = g_array_sized_new(FALSE, FALSE, sizeof(int),
+				     device->styli->len);
+	for (guint i = 0; i < device->styli->len; i++) {
+		int id = g_array_index(device->styli, int, i);
+		g_array_append_val(d->styli, id);
+	}
+	d->status_leds = g_array_sized_new(FALSE, FALSE,
+					   sizeof(WacomStatusLEDs),
+					   device->status_leds->len);
+	for (guint i = 0; i < device->status_leds->len; i++) {
+		g_array_append_val(d->status_leds,
+				   g_array_index(device->status_leds, WacomStatusLEDs, i));
+	}
+
+	d->buttons = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+					   NULL, g_free);
+	g_hash_table_iter_init(&iter, device->buttons);
+	while (g_hash_table_iter_next(&iter, &k, &v)) {
+		WacomButton *a = v;
+		WacomButton *b = g_memdup2(a, sizeof(WacomButton));
+		g_hash_table_insert(d->buttons, k, b);
+	}
 	return d;
 }
 
+static bool
+match_is_equal(const WacomMatch *a, const WacomMatch *b)
+{
+	return g_str_equal(a->match, b->match);
+}
 
-static int
-compare_matches(const WacomDevice *a, const WacomDevice *b)
+static bool
+matches_are_equal(const WacomDevice *a, const WacomDevice *b)
 {
 	const WacomMatch **ma, **mb, **match_a, **match_b;
 
@@ -355,14 +400,14 @@ compare_matches(const WacomDevice *a, const WacomDevice *b)
 	for (match_a = ma; *match_a; match_a++) {
 		int found = 0;
 		for (match_b = mb; !found && *match_b; match_b++) {
-			if (streq((*match_a)->match, (*match_b)->match))
+			if (match_is_equal(*match_a, *match_b))
 				found = 1;
 		}
 		if (!found)
-			return 1;
+			return false;
 	}
 
-	return 0;
+	return true;
 }
 
 /* Compare layouts based on file name, stripping the full path */
@@ -394,12 +439,18 @@ libwacom_same_layouts (const WacomDevice *a, const WacomDevice *b)
 LIBWACOM_EXPORT int
 libwacom_compare(const WacomDevice *a, const WacomDevice *b, WacomCompareFlags flags)
 {
+	GHashTableIter iter;
+	gpointer k, v;
+
 	g_return_val_if_fail(a || b, 0);
 
 	if (!a || !b)
 		return 1;
 
-	if (!streq(a->name, b->name))
+	if (a == b)
+		return 0;
+
+	if (!g_str_equal(a->name, b->name))
 		return 1;
 
 	if (a->width != b->width || a->height != b->height)
@@ -429,36 +480,40 @@ libwacom_compare(const WacomDevice *a, const WacomDevice *b, WacomCompareFlags f
 	if (a->ring2_num_modes != b->ring2_num_modes)
 		return 1;
 
-	if (a->num_buttons != b->num_buttons)
+	if (g_hash_table_size(a->buttons) != g_hash_table_size(b->buttons))
 		return 1;
 
-	if (a->num_styli != b->num_styli)
+	if (a->styli->len != b->styli->len)
 		return 1;
 
-	if (memcmp(a->supported_styli, b->supported_styli, sizeof(int) * a->num_styli) != 0)
+	if (memcmp(a->styli->data, b->styli->data, sizeof(int) * a->styli->len) != 0)
 		return 1;
 
-	if (a->num_leds != b->num_leds)
+	if (a->status_leds->len != b->status_leds->len)
 		return 1;
 
-	if (memcmp(a->status_leds, b->status_leds, sizeof(WacomStatusLEDs) * a->num_leds) != 0)
+	if (memcmp(a->status_leds->data, b->status_leds->data,
+		   sizeof(WacomStatusLEDs) * a->status_leds->len) != 0)
 		return 1;
 
-	if (memcmp(a->buttons, b->buttons, sizeof(WacomButtonFlags) * a->num_buttons) != 0)
-		return 1;
+	g_hash_table_iter_init(&iter, a->buttons);
+	while (g_hash_table_iter_next(&iter, &k, &v)) {
+		WacomButton *ba = v;
+		WacomButton *bb = g_hash_table_lookup(b->buttons, k);
 
-	if (memcmp(a->button_codes, b->button_codes, sizeof(int) * a->num_buttons) != 0)
-		return 1;
+		if (!bb || ba->flags != bb->flags || ba->code != bb->code)
+			return 1;
+	}
 
 	if ((a->paired == NULL && b->paired != NULL) ||
 	    (a->paired != NULL && b->paired == NULL) ||
-	    (a->paired && b->paired && !streq(a->paired->match, b->paired->match)))
+	    (a->paired && b->paired && !match_is_equal(a->paired, b->paired)))
 		return 1;
 
-	if ((flags & WCOMPARE_MATCHES) && compare_matches(a, b) != 0)
-		return 1;
-	else if (!streq(a->matches[a->match]->match, b->matches[b->match]->match))
+	if ((flags & WCOMPARE_MATCHES) && !matches_are_equal(a, b))
 		return 1;
+	else if (!match_is_equal(a->match, b->match))
+			return 1;
 
 	return 0;
 }
@@ -520,15 +575,15 @@ libwacom_new_from_path(const WacomDeviceDatabase *db, const char *path, WacomFal
 		match_name = NULL;
 		device = libwacom_new (db, match_name, vendor_id, product_id, bus, error);
 	}
-	if (device != NULL)
-		ret = libwacom_copy(device);
-	else if (fallback == WFALLBACK_NONE)
-		goto bail;
 
-	if (device == NULL && fallback == WFALLBACK_GENERIC) {
+	if (device == NULL) {
+		if (fallback == WFALLBACK_NONE)
+			goto out;
+
+		/* WFALLBACK_GENERIC */
 		device = libwacom_get_device(db, "generic");
 		if (device == NULL)
-			goto bail;
+			goto out;
 
 		ret = libwacom_copy(device);
 
@@ -536,27 +591,24 @@ libwacom_new_from_path(const WacomDeviceDatabase *db, const char *path, WacomFal
 			g_free (ret->name);
 			ret->name = g_strdup(name);
 		}
+	} else {
+		ret = libwacom_copy(device);
 	}
 
 	/* for multiple-match devices, set to the one we requested */
 	match = libwacom_match_new(match_name, bus, vendor_id, product_id);
-	libwacom_add_match(ret, match);
+	libwacom_set_default_match(ret, match);
 	libwacom_match_unref(match);
 
-	if (device) {
-		/* if unset, use the kernel flags. Could be unset as well. */
-		if (ret->integration_flags == WACOM_DEVICE_INTEGRATED_UNSET)
-			ret->integration_flags = integration_flags;
+	/* if unset, use the kernel flags. Could be unset as well. */
+	if (device && ret->integration_flags == WACOM_DEVICE_INTEGRATED_UNSET)
+		ret->integration_flags = integration_flags;
 
-		g_free (name);
-
-		return ret;
-	}
-
-bail:
+out:
 	g_free (name);
-	libwacom_error_set(error, WERROR_UNKNOWN_MODEL, NULL);
-	return NULL;
+	if (ret == NULL)
+		libwacom_error_set(error, WERROR_UNKNOWN_MODEL, "unknown model");
+	return ret;
 }
 
 LIBWACOM_EXPORT WacomDevice*
@@ -600,7 +652,7 @@ libwacom_new_from_name(const WacomDeviceDatabase *db, const char *name, WacomErr
 	for (l = keys; l; l = l->next) {
 		WacomDevice *d = l->data;
 
-		if (streq(d->name, name)) {
+		if (g_str_equal(d->name, name)) {
 			device = d;
 			break;
 		}
@@ -848,8 +900,6 @@ libwacom_ref(WacomDevice *device)
 WacomDevice *
 libwacom_unref(WacomDevice *device)
 {
-	int i;
-
 	if (device == NULL)
 		return NULL;
 
@@ -863,13 +913,14 @@ libwacom_unref(WacomDevice *device)
 	g_free (device->layout);
 	if (device->paired)
 		libwacom_match_unref(device->paired);
-	for (i = 0; i < device->nmatches; i++)
-		libwacom_match_unref(device->matches[i]);
-	g_free (device->matches);
-	g_free (device->supported_styli);
-	g_free (device->status_leds);
-	g_free (device->buttons);
-	g_free (device->button_codes);
+	for (guint i = 0; i < device->matches->len; i++)
+		libwacom_match_unref(g_array_index(device->matches, WacomMatch*, i));
+	g_array_free (device->matches, TRUE);
+	libwacom_match_unref(device->match);
+	g_array_free (device->styli, TRUE);
+	g_array_free (device->status_leds, TRUE);
+	if (device->buttons)
+		g_hash_table_destroy (device->buttons);
 	g_free (device);
 
 	return NULL;
@@ -891,7 +942,8 @@ libwacom_match_ref(WacomMatch *match)
 WacomMatch*
 libwacom_match_unref(WacomMatch *match)
 {
-	if (!g_atomic_int_dec_and_test(&match->refcnt))
+	if (match == NULL ||
+	    !g_atomic_int_dec_and_test(&match->refcnt))
 		return NULL;
 
 	g_free (match->match);
@@ -926,30 +978,63 @@ libwacom_match_new(const char *name, WacomBusType bus, int vendor_id, int produc
 void
 libwacom_add_match(WacomDevice *device, WacomMatch *newmatch)
 {
-	int i;
+	for (guint i = 0; i < device->matches->len; i++) {
+		WacomMatch *m = g_array_index(device->matches, WacomMatch *, i);
+		const char *matchstr = libwacom_match_get_match_string(m);
 
-	for (i = 0; i < device->nmatches; i++) {
-		const char *matchstr = libwacom_match_get_match_string(device->matches[i]);
-		if (streq(matchstr, newmatch->match)) {
-			device->match = i;
+		if (g_str_equal(matchstr, newmatch->match)) {
 			return;
 		}
 	}
+	libwacom_match_ref(newmatch);
+	g_array_append_val(device->matches, newmatch);
+}
 
-	device->nmatches++;
+void
+libwacom_set_default_match(WacomDevice *device, WacomMatch *newmatch)
+{
+	for (guint i = 0; i < device->matches->len; i++) {
+		WacomMatch *m = g_array_index(device->matches, WacomMatch *, i);
 
-	device->matches = g_realloc_n(device->matches, device->nmatches + 1, sizeof(WacomMatch*));
-	device->matches[device->nmatches] = NULL;
-	device->matches[device->nmatches - 1] = libwacom_match_ref(newmatch);
-	device->match = device->nmatches - 1;
+		if (match_is_equal(m, newmatch)) {
+			libwacom_match_unref(device->match);
+			device->match = libwacom_match_ref(m);
+			return;
+		}
+	}
+	g_return_if_reached();
+}
+
+void
+libwacom_remove_match(WacomDevice *device, WacomMatch *to_remove)
+{
+	for (guint i= 0; i < device->matches->len; i++) {
+		WacomMatch *m = g_array_index(device->matches, WacomMatch*, i);
+		if (match_is_equal(m, to_remove)) {
+			WacomMatch *dflt = device->match;
+
+			/* remove from list */
+			g_array_remove_index(device->matches, i);
+
+			/* now reset the default match if needed */
+			if (match_is_equal(dflt, to_remove)) {
+				WacomMatch *first = g_array_index(device->matches,
+							      WacomMatch*,
+							      0);
+				libwacom_set_default_match(device, first);
+			}
+
+			libwacom_match_unref(to_remove);
+			break;
+		}
+	}
 }
 
 LIBWACOM_EXPORT int
 libwacom_get_vendor_id(const WacomDevice *device)
 {
-	g_return_val_if_fail(device->match >= 0, -1);
-	g_return_val_if_fail(device->match < device->nmatches, -1);
-	return device->matches[device->match]->vendor_id;
+	g_return_val_if_fail(device->match, -1);
+	return device->match->vendor_id;
 }
 
 LIBWACOM_EXPORT const char*
@@ -973,23 +1058,21 @@ libwacom_get_layout_filename(const WacomDevice *device)
 LIBWACOM_EXPORT int
 libwacom_get_product_id(const WacomDevice *device)
 {
-	g_return_val_if_fail(device->match >= 0, -1);
-	g_return_val_if_fail(device->match < device->nmatches, -1);
-	return device->matches[device->match]->product_id;
+	g_return_val_if_fail(device->match, -1);
+	return device->match->product_id;
 }
 
 LIBWACOM_EXPORT const char*
 libwacom_get_match(const WacomDevice *device)
 {
-	g_return_val_if_fail(device->match >= 0, NULL);
-	g_return_val_if_fail(device->match < device->nmatches, NULL);
-	return device->matches[device->match]->match;
+	g_return_val_if_fail(device->match, NULL);
+	return device->match->match;
 }
 
 LIBWACOM_EXPORT const WacomMatch**
 libwacom_get_matches(const WacomDevice *device)
 {
-	return (const WacomMatch**)device->matches;
+	return (const WacomMatch**)device->matches->data;
 }
 
 LIBWACOM_EXPORT const WacomMatch*
@@ -1031,14 +1114,14 @@ libwacom_has_touch(const WacomDevice *device)
 LIBWACOM_EXPORT int
 libwacom_get_num_buttons(const WacomDevice *device)
 {
-	return device->num_buttons;
+	return g_hash_table_size(device->buttons);
 }
 
 LIBWACOM_EXPORT const int *
 libwacom_get_supported_styli(const WacomDevice *device, int *num_styli)
 {
-	*num_styli = device->num_styli;
-	return device->supported_styli;
+	*num_styli = device->styli->len;
+	return (const int *)device->styli->data;
 }
 
 LIBWACOM_EXPORT int
@@ -1080,11 +1163,11 @@ libwacom_get_strips_num_modes(const WacomDevice *device)
 LIBWACOM_EXPORT const WacomStatusLEDs *
 libwacom_get_status_leds(const WacomDevice *device, int *num_leds)
 {
-	*num_leds = device->num_leds;
-	return device->status_leds;
+	*num_leds = device->status_leds->len;
+	return (const WacomStatusLEDs*)device->status_leds->data;
 }
 
-struct {
+static const struct {
 	WacomButtonFlags button_flags;
 	WacomStatusLEDs  status_leds;
 } button_status_leds[] = {
@@ -1097,25 +1180,21 @@ struct {
 LIBWACOM_EXPORT int
 libwacom_get_button_led_group (const WacomDevice *device, char button)
 {
-	int button_index, led_index;
-	WacomButtonFlags button_flags;
+	WacomButton *b = g_hash_table_lookup(device->buttons,
+					     GINT_TO_POINTER(button));
 
-	g_return_val_if_fail (device->num_buttons > 0, -1);
-	g_return_val_if_fail (button >= 'A', -1);
-	g_return_val_if_fail (button < 'A' + device->num_buttons, -1);
-
-	button_index = button - 'A';
-	button_flags = device->buttons[button_index];
-
-	if (!(button_flags & WACOM_BUTTON_MODESWITCH))
+	if (!(b->flags & WACOM_BUTTON_MODESWITCH))
 		return -1;
 
-	for (led_index = 0; led_index < device->num_leds; led_index++) {
+	for (guint led_index = 0; led_index < device->status_leds->len; led_index++) {
 		guint n;
 
 		for (n = 0; n < G_N_ELEMENTS (button_status_leds); n++) {
-			if ((button_flags & button_status_leds[n].button_flags) &&
-			    (device->status_leds[led_index] == button_status_leds[n].status_leds)) {
+			WacomStatusLEDs led = g_array_index(device->status_leds,
+							    WacomStatusLEDs,
+							    led_index);
+			if ((b->flags & button_status_leds[n].button_flags) &&
+			    (led == button_status_leds[n].status_leds)) {
 				return led_index;
 			}
 		}
@@ -1155,37 +1234,27 @@ libwacom_get_integration_flags (const WacomDevice *device)
 LIBWACOM_EXPORT WacomBusType
 libwacom_get_bustype(const WacomDevice *device)
 {
-	g_return_val_if_fail(device->match >= 0, -1);
-	g_return_val_if_fail(device->match < device->nmatches, -1);
-	return device->matches[device->match]->bus;
+	g_return_val_if_fail(device->match, -1);
+	return device->match->bus;
 }
 
 LIBWACOM_EXPORT WacomButtonFlags
 libwacom_get_button_flag(const WacomDevice *device, char button)
 {
-	int index;
+	WacomButton *b = g_hash_table_lookup(device->buttons,
+					     GINT_TO_POINTER(button));
 
-	g_return_val_if_fail (device->num_buttons > 0, WACOM_BUTTON_NONE);
-	g_return_val_if_fail (button >= 'A', WACOM_BUTTON_NONE);
-	g_return_val_if_fail (button < 'A' + device->num_buttons, WACOM_BUTTON_NONE);
 
-	index = button - 'A';
-
-	return device->buttons[index];
+	return b ? b->flags : WACOM_BUTTON_NONE;
 }
 
 LIBWACOM_EXPORT int
 libwacom_get_button_evdev_code(const WacomDevice *device, char button)
 {
-	int index;
+	WacomButton *b = g_hash_table_lookup(device->buttons,
+					     GINT_TO_POINTER(button));
 
-	g_return_val_if_fail (device->num_buttons > 0, 0);
-	g_return_val_if_fail (button >= 'A', 0);
-	g_return_val_if_fail (button < 'A' + device->num_buttons, 0);
-
-	index = button - 'A';
-
-	return device->button_codes[index];
+	return b ? b->code : 0;
 }
 
 LIBWACOM_EXPORT const
@@ -1210,8 +1279,8 @@ LIBWACOM_EXPORT const int *
 libwacom_stylus_get_paired_ids(const WacomStylus *stylus, int *num_paired_ids)
 {
 	if (num_paired_ids)
-		*num_paired_ids = stylus->num_ids;
-	return stylus->paired_ids;
+		*num_paired_ids = stylus->paired_ids->len;
+	return (const int*)stylus->paired_ids->data;
 }
 
 LIBWACOM_EXPORT int
@@ -1345,7 +1414,8 @@ libwacom_stylus_unref(WacomStylus *stylus)
 
 	g_free (stylus->name);
 	g_free (stylus->group);
-	g_free (stylus->paired_ids);
+	if (stylus->paired_ids)
+		g_array_free (stylus->paired_ids, TRUE);
 	g_free (stylus);
 
 	return NULL;
