@@ -24,9 +24,7 @@
  *        Peter Hutterer <peter.hutterer@redhat.com>
  */
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #define _GNU_SOURCE
 #include <stdio.h>
@@ -39,29 +37,31 @@
 static void
 print_device_info(const WacomDeviceDatabase *db, const WacomDevice *device)
 {
-    const int *styli;
-    int nstyli;
+	const int *styli;
+	int nstyli;
 
-    printf("%s", libwacom_get_name(device));
-    if (libwacom_get_model_name(device)) {
-	    printf(" (%s)", libwacom_get_model_name(device));
-    }
-    printf(":\n");
+	printf("- name: '%s'\n", libwacom_get_name(device));
+	if (libwacom_get_model_name(device)) {
+		printf("  model: '%s'\n", libwacom_get_model_name(device));
+	}
+	if (!libwacom_has_stylus(device)) {
+		printf("  styli: []  # no styli defined\n");
+		return;
+	}
 
-    if (!libwacom_has_stylus(device)) {
-        printf("\tno styli defined\n");
-        return;
-    }
+	printf("  styli:\n");
 
-    styli = libwacom_get_supported_styli(device, &nstyli);
-    for (int i = 0; i < nstyli; i++) {
-        const WacomStylus *s;
+	styli = libwacom_get_supported_styli(device, &nstyli);
+	for (int i = 0; i < nstyli; i++) {
+		const WacomStylus *s;
+		char id[64];
 
-        s = libwacom_stylus_get_for_id(db, styli[i]);
-        printf("\t%#8x:\t%s\n",
-               libwacom_stylus_get_id(s),
-               libwacom_stylus_get_name(s));
-    }
+		s = libwacom_stylus_get_for_id(db, styli[i]);
+		snprintf(id, sizeof(id), "0x%x", libwacom_stylus_get_id(s));
+		printf("    - { id: %*s'%s', name: '%s' }\n",
+		       (int)(7 - strlen(id)), " ", id,
+		       libwacom_stylus_get_name(s));
+	}
 }
 
 int main(int argc, char **argv)
@@ -70,12 +70,12 @@ int main(int argc, char **argv)
 	WacomDevice **list, **p;
 
 	if (argc > 1) {
-		printf("Usage: %s [--help] - list all supported devices\n",
+		printf("Usage: %s [--help] - list compatible styli\n",
 		       basename(argv[0]));
 	       return !!(strcmp(argv[1], "--help"));
 	}
 
-	db = libwacom_database_new_for_path(TOPSRCDIR"/data");
+	db = libwacom_database_new_for_path(DATABASEPATH);
 
 	list = libwacom_list_devices_from_database(db, NULL);
 	if (!list) {
